import asyncio
import json
from asyncio import Future
from pathlib import Path
from typing import List, Type, TypeVar, Generic

from pydantic import BaseModel as PdModel
from pyee import AsyncIOEventEmitter

from web_foundation.app.service.service import Service
from web_foundation.kernel import NamedContext
from web_foundation.kernel.isolates.isolate import Isolate

AppConfig = TypeVar("AppConfig", bound=PdModel)


class App(Generic[AppConfig]):
    emitter: AsyncIOEventEmitter
    config: AppConfig
    name: str
    isolates: List[Isolate]
    services: List[Service]
    ctx: NamedContext
    debug: bool

    def __init__(self, name: str, debug: bool = False):
        self.debug = debug
        self.emitter = AsyncIOEventEmitter()
        self.name = name
        self.ctx = NamedContext()
        self.services = []
        self.isolates = []

    def load_config(self, conf_path: Path, config_model: Type[AppConfig]) -> None:
        """
        Load app config to user in
        :param conf_path:
        :param config_model: BaseModel to cast json file to pydantic
        :return: None
        """
        with open(conf_path, "r") as _json_file:
            conf = config_model(**json.loads(_json_file.read()))
            self.config = conf
            self.ctx.app_config = conf

    def add_isolate(self, isolate: Isolate):
        """
        Set isolate to app and set isolate debug and name
        :param isolate: isolate to apply
        :return: None
        """
        if not isolate.configured:
            raise RuntimeError("Isolate not configured")
        isolate.debug = self.debug
        isolate.ctx = self.ctx
        self.isolates.append(isolate)

    def add_service(self, service: Service):
        """
        Set service to use in app
        Set service to named contex
        :param service: service to set
        :return: None
        """
        self.ctx.set_obj(service, kind=Service)
        self.services.append(service)

    async def prepare_infrastructure(self):
        """
        example:
            initor = Provider.of(DatabaseInitiator, self.name)
            for service in self.services:
                await service.prepare(**kwargs : Dict[str, Initiator])
            self.emitter.emit("finalize_init")
        """
        pass

    def perform(self) -> List[Future]:
        """
        Call performs from isolates and return all isolates Futures
        :return: None
        """
        return [isolate.perform() for isolate in self.isolates]

    async def run(self):
        """
        Func to run app manually (without Runner)
        :return: None
        """
        return await asyncio.wait([isolate.perform() for isolate in self.isolates])



from typing import Type

from sanic import Request
from pydantic import ValidationError as PdValidationError, BaseModel as PdModel
from .erorrs.application import InconsistencyError


def validate_dto(request: Request, dto_cls: Type[PdModel]) -> PdModel:
    try:
        dto = dto_cls(**request.json)
        return dto
    except PdValidationError as ex:
        raise InconsistencyError(message=dto_validation_error_format(ex))


def dto_validation_error_format(exeption: PdValidationError):
    failed_fields = exeption.errors()
    commment_str = "Some of essential params failed : " + ", ".join(
        [str(field["loc"][-1]) + " - " + field["msg"] for field in failed_fields])
    return commment_str



import asyncio
from asyncio import Task
from functools import singledispatchmethod

from web_foundation.workers.background.message import TaskIMessage, TaskReplyIMessage
from web_foundation.workers.worker import Worker


class TaskExecutor(Worker):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @singledispatchmethod
    def _configure(self, name: str, *args, **kwargs):
        pass

    @_configure.register
    def _(self, something:str):
        pass

    async def _run(self):
        await self.channel.listen_consume(self.task_handler)

    async def task_handler(self, message: TaskIMessage):
        task = asyncio.create_task(self.exec(message))
        if message.respond:
            task.add_done_callback(self.produce_response)

    async def exec(self, msg: TaskIMessage) -> TaskReplyIMessage:
        return TaskReplyIMessage(msg, await msg.task(msg), self.name)

    def produce_response(self, resp: Task):
        asyncio.create_task(self.channel.produce(resp.result()))

    def _close(self, *args, **kwargs):
        pass



import asyncio
import os
import pathlib
import warnings
from signal import signal as signal_func, SIGINT, SIGTERM, Signals
from typing import List

from loguru import logger
from pydantic import BaseModel as PDModel

from web_foundation.app import App
from web_foundation.app import AbstractDbModel
from web_foundation.app import AccessService, AccessSpec
from web_foundation.app import Service
from web_foundation.app.infrastructure.database.initiator import DbConfig
from web_foundation.kernel import IDistributor
from web_foundation.workers import TaskIMessage, TaskExecutor, PipelinedRouteConf
from web_foundation.workers import ServerConfig, WebServer
from web_foundation.workers import set_to_pipline, Pipeline, mark_pipelined, Protector
from web_foundation.workers import HTTPMethod, ProtectIdentity, InputContext

warnings.filterwarnings("ignore")  # TODO apply to settings


class AppConfg(PDModel):
    server: ServerConfig
    db_config: DbConfig


class AppBaseService(Service):
    pass


class AnyAppPipeline(Pipeline):
    routes_confs: List[PipelinedRouteConf] = []


class AnyAppAccessPipline(Pipeline):
    routes_confs: List[PipelinedRouteConf] = []


@mark_pipelined(AnyAppPipeline)
class AnyService(Service):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.q = "asgajsdhbngaksngkp;amsg"

    async def test(self, msg: TaskIMessage):
        await asyncio.sleep(1)
        logger.warning(self.q)
        logger.warning(msg.args)

    @set_to_pipline("/test/<id:int>", HTTPMethod.GET, AnyAppPipeline)
    async def person_details(self, inc: InputContext, prot: ProtectIdentity):
        # logger.info(inc.r_attrs)
        # logger.warning(inc.channel)
        # logger.warning(inc.named_ctx.get_obj(AnyService))
        await inc.channel.produce(TaskIMessage("test_app_task_executor", self.test, args=("asfasf",)))
        return {}


@mark_pipelined(AnyAppAccessPipline)
class OverrideUserAccess(Service):
    @set_to_pipline("/user/<entity_id:int>", HTTPMethod.GET, AnyAppAccessPipline)
    async def test(self, inc: InputContext, prot: ProtectIdentity):
        """
        test of override accesses
        :param inc:
        :param prot:
        :return:
        """
        return {}


class AnyModel(AbstractDbModel):
    pass


async def main():
    app = App("test_app", debug=True)
    app.load_config(pathlib.Path("./config.json"), AppConfg)
    sock = WebServer.create_socket(app.config.server)

    class Ansss(PDModel):
        pass

    acc = AccessService()
    map = [
        AccessSpec("/user", {HTTPMethod.GET, HTTPMethod.POST}, Protector, AnyModel, Ansss, Ansss),
        AccessSpec("/customer",
                   {HTTPMethod.GET, HTTPMethod.POST}, Protector, AnyModel, Ansss, Ansss)
    ]
    acc.add_default_routes(map, AnyAppAccessPipline)
    app.add_service(acc)
    app.add_service(AnyService())
    app.add_service(OverrideUserAccess())
    # app.add_service(acc)

    # bg_wr = BGWorker("bg_worker")
    # app.add_isolate(bg_wr)
    # for i in range(os.cpu_count()):
    #     web_server = WebIsolate(f"{app.name}_sanic_{i}", app.config.server, sock)
    #     web_server.apply_named_ctx(app.ctx)
    #     app.add_isolate(web_server)
    web_server = WebServer(app.config.server)
    web_server.configure(f"{app.name}_sanic_{1}", app.ctx, sock)
    task = TaskExecutor()
    task.configure(f"{app.name}_task_executor", "something")
    web_server.add_pipelines(AnyAppPipeline)
    web_server.add_pipelines(AnyAppAccessPipline)
    app.add_isolate(web_server)
    app.add_isolate(task)
    distrib = IDistributor(app.isolates)

    def sig_handler(signal, frame):
        logger.info("Received signal %s. Shutting down.", Signals(signal).name)
        for i in app.isolates:
            i.process.terminate()
        sock.close()

    signal_func(SIGINT, lambda s, f: sig_handler(s, f))
    signal_func(SIGTERM, lambda s, f: sig_handler(s, f))

    await asyncio.wait(app.perform() + distrib.preform())

    # sock.close()

    # signal_func(SIGBREAK, lambda s, f: sig_handler(s, f))


if __name__ == '__main__':
    asyncio.run(main())



import asyncio
from typing import Coroutine, Callable, TypeVar

from aioprocessing import AioQueue, AioEvent, AioLock
from loguru import logger

from web_foundation.kernel.isolates.pipes import IsolatePipes, IMessage, GenericIMessage


class IChannel:
    read_timeout = 0.01
    pipes: IsolatePipes
    name: str
    debug: bool
    consume_pipe: IsolatePipes
    produce_pipe: IsolatePipes

    def __init__(self, wroker_name: str, debug: bool = False):
        self.name = wroker_name
        self.debug = debug
        self.produce_pipe = IsolatePipes(AioQueue(), AioEvent(), AioLock())
        self.consume_pipe = IsolatePipes(AioQueue(), AioEvent(), AioLock())

    def __str__(self):
        return f"IChannel(name={self.name})"

    async def produce(self, msg: IMessage):
        msg.sender = self.name
        logger.warning(msg)
        await self.produce_pipe.write(msg)

    async def _listen(self, pipe: IsolatePipes, callback: Callable[[GenericIMessage], Coroutine]):
        while True:
            if pipe.empty():
                await asyncio.sleep(self.read_timeout)
            r: GenericIMessage = await pipe.read()
            if self.debug:
                logger.info(f"Channel {self.name} {'send' if pipe == self.produce_pipe else 'receive'} message: {r}")
            await callback(r)

    async def listen_produce(self, callback: Callable[[GenericIMessage], Coroutine]):
        await self._listen(self.produce_pipe, callback)

    async def listen_consume(self, callback: Callable[[GenericIMessage], Coroutine]):
        await self._listen(self.consume_pipe, callback)



from __future__ import annotations

from typing import Dict, Type, TypeVar, Generic

T = TypeVar("T")


class Provider(Generic[T]):
    provided_entities: Dict[str, Dict[Type[T], T]] = {}

    @staticmethod
    def reg(app_name: str, initiator):
        Provider.provided_entities.update({app_name: {type(initiator): initiator}})  # type: ignore

    @staticmethod
    def of(initiator_type: Type[T], app_name: str) -> T:
        if not Provider.provided_entities.get(app_name):  # type: ignore
            raise ValueError(f"App {app_name} not found in provide context")
        if not Provider.provided_entities.get(app_name).get(initiator_type):  # type: ignore
            raise ValueError(f"Provider {initiator_type.__name__} for {app_name} not found in provide context")
        return Provider.provided_entities.get(app_name).get(initiator_type)  # type: ignore



from typing import TypeVar, Generic

from web_foundation.app import App
from web_foundation.app.infrastructure.provide.provider import Provider
from web_foundation.kernel import NamedContext

InitableApp = TypeVar('InitableApp', bound=App, contravariant=True)


class Initiator(Generic[InitableApp]):
    app: InitableApp
    close_event_name = "finalize_init"

    def __init__(self, app: InitableApp):
        self.app = app
        Provider.reg(app.name, self)
        pass

    async def close(self):
        raise NotImplementedError

    async def setup_connection(self, named_ctx: NamedContext | None = None):
        raise NotImplementedError



from .app import App
from .service.service import Service
from .service.access import AccessService, AccessSpec
from .infrastructure.database.models import AbstractDbModel
from .infrastructure.database import utils as db_utils
from .infrastructure.database.middleware import AccessDMW, RelateDMW, DatabaseMiddleware
from .infrastructure.provide.initiator import Initiator
from .infrastructure.database.initiator import DatabaseInitiator
from .infrastructure.provide.provider import Provider

__all__ = ("App",
           "Service",
           "AccessService",
           "AccessSpec",
           "AbstractDbModel",
           "db_utils",
           "AccessDMW",
           "RelateDMW",
           "DatabaseMiddleware",
           "Initiator",
           "DatabaseInitiator",
           "Provider"
           )



import warnings
from dataclasses import dataclass
from distutils.version import LooseVersion
from functools import wraps
from types import MethodType, SimpleNamespace
from typing import List, Any, Dict, Callable, Type, TypeVar, Generic, ParamSpec, ClassVar

from mypy_extensions import VarArg, KwArg
from pydantic import BaseModel as PdModel
from sanic import __version__ as sanic_version, Request, json, text, HTTPResponse

from web_foundation.kernel import class_of_method, NamedContext
from web_foundation.workers.web.sanic_ext.structs import InputContext, ProtectIdentity, HTTPMethod
from web_foundation.workers.web.utils import validate_dto

try:
    import sanic_ext
    from sanic_ext.extensions.base import Extension
    from sanic_ext.config import Config

    use_ext = True
except ImportError:
    use_ext = False
    Extension = object
    from sanic.config import Config
SANIC_VERSION = LooseVersion(sanic_version)
SANIC_21_9_0 = LooseVersion("21.9.0")

P = ParamSpec("P")


class RequestParser:
    extracted_attrs: List[str] | None
    struct: PdModel | None
    dto_validation_fnc: Callable[..., PdModel]

    def __init__(self,
                 dto_validation_fnc: Callable[..., PdModel],
                 extracted_attrs: List[str] = None,
                 struct: PdModel = None):
        self.dto_validation_fnc = dto_validation_fnc  # type: ignore
        self.struct = struct
        if extracted_attrs:
            for attr in extracted_attrs:
                if not hasattr(Request, attr):
                    raise AttributeError(f"sanic.Request has not attribute {attr}")
        self.extracted_attrs = extracted_attrs

    def income_(self, request: Request, **kwargs) -> InputContext:
        incoming_context = InputContext(r_args=SimpleNamespace(), r_kwargs={}, named_ctx=request.app.ctx.named_ctx,
                                        dto=None, channel=request.app.ctx.channel, request=request)
        if self.extracted_attrs:
            for attr in self.extracted_attrs:
                setattr(incoming_context.r_args, attr, getattr(request, attr))
        incoming_context.r_kwargs = kwargs
        if self.struct:
            incoming_context.dto = self.dto_validation_fnc(request, self.struct)
        return incoming_context


class ResponseConstructor:
    fabric: Callable[[Any], HTTPResponse] | None
    default_response: HTTPResponse
    pd_model: Type[PdModel] | None

    def __init__(self,
                 pd_model: Type[PdModel] = None,
                 fabric: Callable[[Any], HTTPResponse] = None,
                 default_response: HTTPResponse = text("OK")):
        self.default_response = default_response
        self.pd_model = pd_model
        self.fabric = fabric

    def return_(self, to_response: Any) -> HTTPResponse:
        if self.fabric:
            return self.fabric(to_response)
        else:
            return self.default_response


class Protector:
    named_ctx: NamedContext

    def __init__(self, named_ctx: NamedContext):
        self.named_ctx = named_ctx

    @classmethod
    async def protect(cls, request: Request) -> ProtectIdentity | None:
        return None


GenericPipline = TypeVar("GenericPipline")


@dataclass()
class PipelinedRouteConf(Generic[GenericPipline]):
    pipeline: GenericPipline | List[GenericPipline] | None
    uri: str
    method: HTTPMethod
    target: Callable[[VarArg(Any), KwArg(Any)], Any]
    protector: Type[Protector] | None
    kwargs: Dict[str, Any]

    def __eq__(self, other):
        return self.uri == other.uri and self.method == other.method

    def __str__(self):
        return f"Route-conf in pipeline {self.pipeline} to {self.uri} {self.method} on {self.target}"


class Pipeline:
    routes_confs: List[PipelinedRouteConf]

    def __str__(self):
        return ";".join(str(self.routes_confs))


class PipelinesExt(Extension):
    name: str = "pipelines"
    _pipelines: List[Pipeline]

    def __init__(self):
        self._pipelines = []

    @property
    def pipelines(self):
        return self._pipelines

    @pipelines.setter
    def pipelines(self, value: List[Pipeline]):
        self._pipelines = value

    def startup(self, bootstrap) -> None:
        if SANIC_21_9_0 > SANIC_VERSION:
            raise RuntimeError(
                "You cannot use this version of SanicPipeline with"
                "Sanic earlier than v21.9.0")
        for pipel in self._pipelines:
            for route_cfg in pipel.routes_confs:
                try:
                    bootstrap.app.ctx.named_ctx.get_obj(class_of_method(route_cfg.target))
                except NamedContext.NotFindInContext:
                    raise AttributeError(f"Can't start unset {class_of_method(route_cfg.target)} to named context")
                bootstrap.app.add_route(route_cfg.target, route_cfg.uri, methods=[route_cfg.method.value],
                                        **route_cfg.kwargs)

    def label(self):
        return f"pipelines added ({', '.join([i.__name__ for i in self._pipelines])})"


def set_to_pipline(uri: str,
                   method: HTTPMethod,
                   pipline: GenericPipline | List[GenericPipline] | None = None,
                   protector: Type[Protector] | None = Protector,
                   struct: Type[PdModel] | None = None,
                   out_struct: Type[PdModel] | None = None,
                   extracted_attrs: List[str] = None,
                   dto_validation_fnc: Callable[..., PdModel] = validate_dto,
                   return_fabric: Callable[[Any], HTTPResponse] = json,
                   default_response: HTTPResponse = text("ok"),
                   **route_kwargs):
    def called_method(target: MethodType):
        @wraps(target)
        async def f(*args, **kwargs):
            req: Request = args[0]
            n_ctx = req.app.ctx.named_ctx
            prot_identity = await (protector(n_ctx).protect(req))
            incoming = RequestParser(dto_validation_fnc, extracted_attrs, struct).income_(req, **kwargs)
            ret_val = await target(n_ctx.get_obj(class_of_method(target)), incoming, prot_identity)
            return ResponseConstructor(out_struct, return_fabric, default_response).return_(ret_val)

        setattr(f, "route_conf", PipelinedRouteConf(pipline, uri, method, f, protector, route_kwargs))
        return f

    return called_method


def mark_pipelined(*args: Type[Pipeline]):
    def wrp(cls):
        for name, method in cls.__dict__.items():
            if hasattr(method, "route_conf"):
                for pipline in args:
                    if isinstance(method.route_conf.pipeline, list):
                        for p in method.route_conf.pipeline:
                            if p == pipline and method.route_conf not in pipline.routes_confs:
                                pipline.routes_confs.append(method.route_conf)
                    elif issubclass(method.route_conf.pipeline, Pipeline):
                        if method.route_conf.pipeline == pipline and method.route_conf not in pipline.routes_confs:
                            pipline.routes_confs.append(method.route_conf)
                    else:
                        warnings.warn(f"Pipline to {method.route_conf} NOT DEFINED")
        return cls

    return wrp



from typing import Type, TypeVar, Dict, Generic

from web_foundation.kernel.pyext.named_obj import NamedObject
from web_foundation.kernel.pyext.namespace import Obj, IterableNamespace, TypedNamespace

CtxKind = TypeVar("CtxKind", bound=NamedObject)


class DefaultKind(NamedObject):
    pass


class NamedContext(Generic[CtxKind], IterableNamespace):
    _spaces: Dict[Type, TypedNamespace]
    _founded_obj: Dict[Type[NamedObject], NamedObject]

    class NotFindInContext(Exception):
        pass

    def __init__(self):
        super(NamedContext, self).__init__()
        self._spaces = {DefaultKind: TypedNamespace()}
        self._founded_obj = {}

    def get_obj(self, obj_type: Type[Obj]) -> NamedObject:
        if obj_type in self._founded_obj:
            return self._founded_obj[obj_type]
        obj = None
        for kind, space in self._spaces.items():
            try:
                obj = space.get_(obj_type)
            except AttributeError as e:
                pass
        if not obj:
            raise self.NotFindInContext(f"Obj type {obj_type} not found in all kinds in context")
        self._founded_obj.update({obj_type: obj})
        return obj

    def set_obj(self, obj: NamedObject, kind: Type = DefaultKind) -> None:
        # kind_aply = kind if kind else DefaultKind
        if kind not in self._spaces:
            self._spaces.update({kind: TypedNamespace()})
        self._spaces[kind].set_(obj)



import asyncio
from asyncio import Future
from functools import singledispatchmethod

from aioprocessing import AioProcess

from web_foundation.kernel import NamedContext
from web_foundation.kernel.isolates.channel import IChannel
from abc import ABCMeta, abstractmethod


class Isolate(metaclass=ABCMeta):
    debug: bool
    _name: str
    _channel: IChannel
    _proc: AioProcess
    _async_mode: bool
    _named_ctx: NamedContext
    _configured: bool

    def __init__(self, async_mode: bool = False, debug: bool = False, configured: bool = False):
        self._async_mode = async_mode
        self.debug = debug
        self._configured = configured

    @property
    def name(self):
        return self._name

    @property
    def ctx(self) -> NamedContext:
        return self._named_ctx

    @ctx.setter
    def ctx(self, value: NamedContext):
        self._named_ctx = value

    @property
    def configured(self) -> bool:
        return self._configured

    @abstractmethod
    async def _run(self):
        pass

    @singledispatchmethod
    def _configure(self, *args, **kwargs):
        pass

    def configure(self, name: str, *args, **kwargs):
        self._name = name
        self._proc = AioProcess(target=self._startup)
        self._channel = IChannel(self._name)
        self._configure(*args, **kwargs)
        self._configured = True

    @abstractmethod
    def _close(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        self._close(*args, **kwargs)

    def _startup(self):
        asyncio.run(self._run())

    async def _exec(self):
        self._proc.start()

    def perform(self) -> Future:
        if not self.configured:
            raise RuntimeError(f"Isolate {self} not configured before start")
        return asyncio.ensure_future(self._exec())

    @property
    def channel(self) -> IChannel:
        return self._channel

    @property
    def pid(self) -> int:
        return self._proc.pid

    @property
    def process(self) -> AioProcess:
        return self._proc



from typing import List

from web_foundation.app.infrastructure.provide.initiator import Initiator
from web_foundation.kernel import Isolate


class Worker(Isolate):
    _initiators: List[Initiator]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initiators = []

    @property
    def initiators(self) -> List[Initiator]:
        return self._initiators

    @initiators.setter
    def initiators(self, value: List[Initiator]):
        self._initiators = value

    async def initiate_on_up(self, initiators: List[Initiator]):
        for init in initiators:
            await init.setup_connection(self.ctx)

    async def _run(self):
        raise NotImplementedError

    def _close(self):
        pass



import asyncio
from asyncio import Task
from typing import Dict, List

from loguru import logger

from web_foundation.kernel import Isolate
from web_foundation.kernel.isolates.channel import IChannel
from web_foundation.kernel.isolates.pipes import GenericIMessage


class IDistributor:
    channels: Dict[str, IChannel]

    def __init__(self, isolates: List[Isolate]):
        self.channels = {}
        for iso in isolates:
            self.channels.update({iso.name: iso.channel})

    async def on_channel_sent(self, msg: GenericIMessage):
        logger.warning(msg)
        channel: IChannel = self.channels[msg.destination]
        await channel.consume_pipe.write(msg)

    def preform(self) -> List[Task]:
        tasks = []
        for channel in self.channels.values():
            tasks.append(asyncio.ensure_future(
                channel.listen_produce(
                    self.on_channel_sent
                )))
        return tasks



from functools import partial
from typing import Callable, Tuple, Dict, Any

from web_foundation.kernel import IMessage
from web_foundation.kernel.isolates.channel import GenericIMessage

BgTask = Callable[[GenericIMessage], Any] | partial


class TaskIMessage(IMessage):
    task: BgTask
    args: Tuple
    kwargs: Dict
    respond: bool

    def __init__(self, destination: str, task: BgTask, args: Tuple[Any] = None, kwargs: Dict[Any, Any] = None,
                 respond: bool = False):
        super().__init__(destination)
        self.task = task
        self.respond = respond
        self.args = args if args else tuple()
        self.kwargs = kwargs if kwargs else {}


class TaskReplyIMessage(IMessage):
    response: Any

    def __init__(self, task_msg: TaskIMessage, response: Any, sender: str):
        super().__init__(task_msg.sender, idx=task_msg.index)
        self.response = response
        self.sender = sender



import random
import string
from typing import Set

from .singleton import Singleton


class NamedObject(Singleton):
    sname: str
    _named: Set[str] = set()

    def __new__(cls, *args, **kwargs):
        super(NamedObject, cls).__new__(cls, *args, **kwargs)
        while True:
            cls.sname = cls._gensname()
            if cls.sname in NamedObject._named:
                continue
            NamedObject._named.add(cls.sname)
            break
        return cls._instance

    @classmethod
    def _gensname(cls) -> str:
        return "".join(random.choices(string.ascii_uppercase + string.digits, k=8))

# if __name__ == '__main__':
#     s = ShadowNamedSpace()
#     t = Test()
#     print(Test.sname)
#     s.set_shadow(t)
#     print(s.get_shadow(Test))
#     print(type(s.__iter__()))
#     for k, v in s:
#         print(k, v)



import asyncio
import os
import socket
from typing import List, Type, Any
from functools import singledispatchmethod
from pydantic import BaseModel as PDModel
from sanic import Sanic
from sanic.server.socket import bind_socket

from web_foundation.kernel import NamedContext
from web_foundation.workers.web.sanic_ext.pipline_ext import Pipeline, PipelinesExt
from web_foundation.workers.worker import Worker


class StreamingConf(PDModel):
    listen_timeout: float
    ping_timeout: float


class ServerConfig(PDModel):
    host: str
    port: int
    streaming: StreamingConf | None


class WebServer(Worker):
    config: ServerConfig
    pipelines_ext: PipelinesExt

    def __init__(self, config: ServerConfig, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = config

    @singledispatchmethod
    def _configure(self, *args, **kwargs):
        pass

    @_configure.register
    def _(self, ctx: NamedContext, sock: socket.socket = None):
        self.socket = sock if sock else self.create_socket(self.config)
        self.sanic_app = Sanic(self.name)
        self.sanic_app.after_server_stop(self.close)
        self.sanic_app.ctx.named_ctx = ctx
        self.sanic_app.ctx.channel = self.channel
        self.pipelines_ext = PipelinesExt()

    def add_pipelines(self, pipelines: List[Type[Pipeline]] | Type[Pipeline]):
        if isinstance(pipelines, list):
            self.pipelines_ext.pipelines = pipelines
        else:
            self.pipelines_ext.pipelines.append(pipelines)

    def _init_ext(self):
        self.sanic_app.extend(extensions=[self.pipelines_ext], config={})

    async def _run(self):
        await self.initiate_on_up(self.initiators)
        self._init_ext()

    def _startup(self):
        super(WebServer, self)._startup()
        try:
            self.sanic_app.run(sock=self.socket)
        except KeyboardInterrupt:
            self._close()

    def _close(self, *args, **kwargs):
        self.socket.close()

    @staticmethod
    def create_socket(config: ServerConfig) -> socket.socket:
        sock = bind_socket(config.host, config.port)
        sock.set_inheritable(True)
        return sock



from dataclasses import dataclass, field
from functools import partial, wraps
from typing import List, Set, Type, Generic, Protocol
import json as py_json
from pydantic import BaseModel as PdModel
from sanic import json
from sanic.exceptions import NotFound

from web_foundation.app.infrastructure.database.middleware import AccessDMW, AccessProtectIdentity  # type: ignore
from web_foundation.app.infrastructure.database.models import AbstractDbModel, GenericDbType
from web_foundation.app.service.service import Service
from web_foundation.workers.web.sanic_ext.pipline_ext import set_to_pipline, Protector, Pipeline, mark_pipelined
from web_foundation.workers.web.sanic_ext.structs import HTTPMethod, ProtectIdentity, InputContext


@dataclass
class AccessSpec(Generic[GenericDbType]):
    uri: str
    methods: Set[HTTPMethod]
    model: GenericDbType
    protector: Type[Protector] | None = field(default=None)
    create_struct: Type[PdModel] | None = field(default=None)
    update_struct: Type[PdModel] | None = field(default=None)


class AccessService(Service):

    def add_default_routes(self, access_mapping: List[AccessSpec], pipline: Type[Pipeline],
                           access_middleware: AccessDMW | None = None):
        db_middleware = access_middleware if access_middleware else AccessDMW
        for acc in access_mapping:
            for method in acc.methods:
                match method:
                    case HTTPMethod.GET:
                        r = set_to_pipline(acc.uri,
                                           method,
                                           pipline,
                                           protector=acc.protector)(
                            wraps(AccessService.exec_access)(partial(AccessService.exec_access,
                                                                     model=acc.model,
                                                                     access_middleware=db_middleware
                                                                     )))
                        setattr(AccessService.exec_access, "route_conf", r.route_conf)
                        mark_pipelined(pipline)(AccessService)
                        r = set_to_pipline(acc.uri + "/<entity_id:int>",
                                           method,
                                           pipline,
                                           protector=acc.protector)(
                            wraps(AccessService.exec_access)(partial(AccessService.exec_access,
                                                                     model=acc.model,
                                                                     access_middleware=db_middleware
                                                                     )))
                        setattr(AccessService.exec_access, "route_conf", r.route_conf)
                        mark_pipelined(pipline)(AccessService)
                    case HTTPMethod.POST:
                        r = set_to_pipline(acc.uri,
                                           method,
                                           pipline,
                                           struct=acc.create_struct,
                                           protector=acc.protector)(
                            wraps(AccessService.exec_access)(partial(AccessService.exec_access,
                                                                     model=acc.model,
                                                                     access_middleware=db_middleware
                                                                     )))
                        setattr(AccessService.exec_access, "route_conf", r.route_conf)
                        mark_pipelined(pipline)(AccessService)
                    case HTTPMethod.PATCH:
                        r = set_to_pipline(acc.uri + "/<entity_id:int>",
                                           method,
                                           pipline,
                                           struct=acc.update_struct,
                                           protector=acc.protector)(
                            wraps(AccessService.exec_access)(partial(AccessService.exec_access,
                                                                     model=acc.model,
                                                                     access_middleware=db_middleware
                                                                     )))
                        setattr(AccessService.exec_access, "route_conf", r.route_conf)
                        mark_pipelined(pipline)(AccessService)
                    case HTTPMethod.DELETE:
                        r = set_to_pipline(acc.uri + "/<entity_id:int>",
                                           method,
                                           pipline,
                                           protector=acc.protector)(
                            wraps(AccessService.exec_access)(partial(AccessService.exec_access,
                                                                     model=acc.model,
                                                                     access_middleware=db_middleware
                                                                     )))
                        setattr(AccessService.exec_access, "route_conf", r.route_conf)
                        mark_pipelined(pipline)(AccessService)

    async def exec_access(self,
                          inc: InputContext,
                          prot: AccessProtectIdentity,
                          model: AbstractDbModel,
                          access_middleware: AccessDMW
                          ):
        """
        Access to models
        :param inc:
        :param prot:
        :param model:
        :return:
        """
        match inc.request.method:
            case HTTPMethod.GET.value:
                kwargs = inc.r_kwargs
                if prot.entity_id is None:
                    limit = inc.request.args.get("limit")
                    offset = inc.request.args.get("offset")
                    limit = int(limit) if limit and limit.isdigit() else 100
                    offset = int(offset) if offset and offset.isdigit() else None
                    # order_by = inc.request.args.get("order_by")  # for openapi. Variable forwarded to read_all in kwargs
                    for ar, val in inc.request.args.items():
                        if ar in ["limit", "offset"]:
                            continue
                        val = val[0]
                        if val.lower() == 'false':
                            val = False
                        elif val.lower() == 'true':
                            val = True
                        elif val.isdigit():
                            val = int(val)
                        elif "." in val and val.replace('.', '').isdigit():
                            val = float(val)
                        elif val.startswith("["):
                            val = py_json.loads(val)
                        kwargs[ar] = val

                    retrieved_all, total = await access_middleware.read_all(model, prot, limit, offset, **kwargs)
                    res = [await model.values_dict() for model in retrieved_all]
                    return res
                retrieved = await access_middleware.read(model, prot, **kwargs)
                if retrieved:
                    result = await retrieved.values_dict()
                    return result
                else:
                    raise NotFound()

            case HTTPMethod.POST.value:
                retrieved = await access_middleware.create(model, prot, inc.dto, **inc.r_kwargs)
                if isinstance(retrieved, list):
                    return json([await i.values_dict() for i in retrieved])
                return json(await retrieved.values_dict())

            case HTTPMethod.PATCH.value:
                retrieved = await access_middleware.update(model, prot, inc.dto, **inc.r_kwargs)
                return json(await retrieved.values_dict())

            case HTTPMethod.DELETE.value:
                retrieved = await access_middleware.delete(model, prot, **inc.r_kwargs)
                return json(await retrieved.values_dict())



from .pyext.singleton import Singleton
from .pyext.additional import class_of_method
from .pyext.named_obj import NamedObject
from .pyext.namespace import IterableNamespace, TypedNamespace
from .pyext.named_context import NamedContext
from .isolates.isolate import Isolate
from .isolates.channel import IChannel
from .isolates.pipes import IMessage, IsolatePipes
from .isolates.distributor import IDistributor

__all__ = ("Singleton",
           "TypedNamespace",
           "IterableNamespace",
           "NamedObject",
           "NamedContext",
           "class_of_method",
           "Isolate",
           "IChannel",
           "IMessage",
           "IsolatePipes",
           "IDistributor",
           )



from dataclasses import dataclass, field
from enum import Enum
from types import SimpleNamespace
from typing import Dict

from pydantic import BaseModel as PdModel
from sanic import Request

from web_foundation.kernel import NamedContext
from web_foundation.kernel.isolates.channel import IChannel


@dataclass
class ProtectIdentity:
    pass


@dataclass
class InputContext:
    r_args: SimpleNamespace
    r_kwargs: Dict
    named_ctx: NamedContext
    channel: IChannel
    dto: PdModel | None
    request: Request


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PATCH = "PATCH"
    PUT = "PUT"
    DELETE = "DELETE"

    @staticmethod
    def all():
        return [HTTPMethod.GET, HTTPMethod.POST, HTTPMethod.PATCH, HTTPMethod.DELETE]



from .background.worker import TaskExecutor
from .background.message import TaskIMessage, TaskReplyIMessage
from .web.sanic_ext.error_handler import ErrorHandlerExt
from .web.sanic_ext.log_handler import LogsHandlerExt
from .web.sanic_ext.pipline_ext import Pipeline, RequestParser, ResponseConstructor, \
    Protector, PipelinedRouteConf, PipelinesExt, set_to_pipline, mark_pipelined
from .web.sanic_ext.structs import InputContext, ProtectIdentity, HTTPMethod
from .web.worker import ServerConfig, WebServer
from .worker import Worker

__all__ = (
    "TaskExecutor",
    "TaskIMessage",
    "TaskReplyIMessage",
    "Pipeline",
    "RequestParser",
    "ResponseConstructor",
    "Protector",
    "PipelinesExt",
    "set_to_pipline",
    "PipelinedRouteConf",
    "mark_pipelined",
    "InputContext",
    "ProtectIdentity",
    "HTTPMethod",
    "LogsHandlerExt",
    "ErrorHandlerExt",
    "ServerConfig",
    "WebServer",
    "Worker"
)



class ApplicationError(Exception):
    def __init__(self, ex=None, message=None):
        if ex:
            self.message = str(ex)
        else:
            self.message = message
        super().__init__(self.message)


class InconsistencyError(ApplicationError):
    def __init__(self, ex=None, message=None):
        super().__init__(ex=ex, message=message)



import sys
from types import MethodType
from typing import Type, Callable, Any

import loguru
from mypy_extensions import VarArg, KwArg

from .namespace import Obj


def class_of_method(method: Callable[[VarArg(Any), KwArg(Any)], Any]) -> Type[Obj]:
    # return method.__self__
    return vars(sys.modules[method.__module__])[method.__qualname__.split('.')[0]]



from dataclasses import dataclass
from typing import TypeVar

import loguru
from aioprocessing import AioEvent, AioLock, AioQueue  # type: ignore


class IMessage:
    last_idx = 0
    index: int
    sender: str
    destination: str

    def __init__(self, destination: str, idx: int = None):
        self.destination = destination
        if not idx:
            IMessage.last_idx += 1
            self.index = IMessage.last_idx
        else:
            self.index = idx

    def __str__(self):
        return f"IMessage(id={self.index}, dest={self.destination}, sender={self.sender})"


GenericIMessage = TypeVar("GenericIMessage", bound=IMessage, contravariant=True)


@dataclass
class IsolatePipes:
    queue: AioQueue
    event: AioEvent
    lock: AioLock

    async def write(self, msg: GenericIMessage):
        await self.queue.coro_put(msg)

    async def read(self) -> GenericIMessage:
        return await self.queue.coro_get()

    def empty(self):
        return self.queue.empty()



from types import SimpleNamespace
from typing import Tuple, Any, Iterator
from typing import Type, TypeVar

from web_foundation.kernel.pyext.named_obj import NamedObject

Obj = TypeVar("Obj", bound=NamedObject, covariant=True)


class IterableNamespace(SimpleNamespace):

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        return iter((item_name, item) for item_name, item in self.__dict__.items())


class TypedNamespace(IterableNamespace):

    def __contains__(self, item: Type[NamedObject]):
        return item in self.__dict__

    def get_(self, item: Type[Obj]) -> Obj:
        return self.__getattribute__(item.sname)

    def set_(self, cls: NamedObject):
        setattr(self, cls.sname, cls)



import os
from pathlib import Path
from types import ModuleType
from typing import List, TypeVar

from loguru import logger
from tortoise import Tortoise, BaseDBAsyncClient, connections

from web_foundation.app.infrastructure.provide.initiator import Initiator, InitableApp
from web_foundation.kernel import NamedContext
from pydantic import BaseModel as PDModel


class DbConfig(PDModel):
    host: str
    port: str
    database: str
    user: str
    password: str
    db_schema: str
    with_migrations: bool
    migrations_path: Path


class DatabaseInitiator(Initiator):
    modules: List[ModuleType]
    connection: BaseDBAsyncClient | None

    def __init__(self, app: InitableApp, modules: List[ModuleType],
                 engine: str = 'tortoise.backends.asyncpg'):
        super(DatabaseInitiator, self).__init__(app)
        self.db_conf = app.config.db_config
        self.modules = modules
        self.connection = None
        self.engine = engine
        self.app.emitter.on(self.close_event_name, self.close)

    def _get_connection_setting(self) -> dict:
        to_discover = [i.__name__ for i in self.modules]
        if self.db_conf.with_migrations:
            to_discover.append("aerich.models")
        return {
            'connections': {
                # Dict format for connection
                f'{self.app.name}_default': {
                    'engine': self.engine,
                    'credentials': {
                        'host': self.db_conf.host,
                        'port': self.db_conf.port,
                        'user': self.db_conf.user,
                        'password': self.db_conf.password,
                        'database': self.db_conf.database,
                        'schema': self.db_conf.db_schema,
                        'minsize': 1,
                        'maxsize': 5,
                    }
                }
            },
            'apps': {
                f'{self.app.name}': {
                    'models': to_discover,
                    'default_connection': f'{self.app.name}_default',
                }
            },
            'use_tz': False,
            'timezone': 'UTC'
        }

    async def fill_db_data(self):
        pass

    async def setup_connection(self, named_ctx: NamedContext | None = None):
        await Tortoise.init(config=self._get_connection_setting())

    @property
    def conn(self) -> BaseDBAsyncClient:
        if not self.connection:
            self.connection = Tortoise.get_connection(connection_name=f'{self.app.name}_default')
        return self.connection

    async def _migrations(self, schema_exists: bool, command):
        path_exists = os.path.exists(os.path.join(os.getcwd(), self.db_conf.migrations_path))

        if not path_exists and not schema_exists:
            await self.create_schema(False)
            await command.init()
            await command.init_db(safe=True)
        elif not schema_exists:
            await self.create_schema(False)
            await command.init()
            await command.upgrade()
        else:
            await command.init()
        logger.info(f"Apply migrations from {self.db_conf.migrations_path}")
        await command.migrate()
        await command.upgrade()

    async def configure_db(self):
        await self.setup_connection()
        row_count, rows = await self.conn.execute_query(
            f"SELECT schema_name FROM information_schema.schemata WHERE schema_name = {self.db_conf.schema}")
        schema_exists = True if row_count else False
        if self.db_conf.with_migrations:
            if not self.db_conf.migrations_path:
                raise ValueError("Field needmigrations_aeirch in db config set to false, can't migrate")
            from aerich import Command
            command = Command(tortoise_config=self._get_connection_setting(), app=self.app.name,
                              location=self.db_conf.migrations_path)
            await self._migrations(schema_exists, command)
        if not schema_exists:
            await self.fill_db_data()

    async def close(self):
        await connections.close_all()

    async def create_schema(self, generate_schemas: bool = True):
        await self.conn.execute_script(f"CREATE SCHEMA IF NOT EXISTS {self.db_conf.schema};")
        if generate_schemas:
            await Tortoise.generate_schemas()
